# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from importlib import import_module
import logging

from ulogger import exceptions


def _setup_default_handler(progname, fmt=None, datefmt=None, **_):
    """Create a Stream handler (default handler).

    Args:
        progname (str): Name of program.
        fmt (:obj:`str`, optional): Desired log format if different than
            the default; uses the same formatting string options
            supported in the stdlib's `logging` module.
        datefmt (:obj:`str`, optional): Desired date format if different
            than the default; uses the same formatting string options
            supported in the stdlib's `logging` module.

    Returns:
        (obj): Instance of `logging.StreamHandler`
    """
    handler = logging.StreamHandler()

    if not fmt:
        # ex: 2017-08-26T14:47:44.968+00:00 <progname> (<PID>) INFO: <msg>
        fmt_prefix = '%(asctime)s.%(msecs)03dZ '
        fmt_suffix = ' (%(process)d) %(levelname)s: ' + '%(message)s'
        fmt = fmt_prefix + progname + fmt_suffix
    if not datefmt:
        datefmt = '%Y-%m-%dT%H:%M:%S'

    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)
    handler.setFormatter(formatter)
    return handler


def setup_logging(progname, level, handlers, **kwargs):
    """Setup logging to stdout (stream), syslog, or stackdriver.

    Attaches handler(s) and sets log level to the root logger.

    Example usage:

        import logging
        from ulogger import setup_logging

        setup_logging('my_awesome_program', 'INFO', ['stream'])

        logging.info('ohai')

    Args:
        progname (str): Name of program.
        level (str): Threshold for when to log.
        handlers (list): Desired handlers, default 'stream',
            supported: 'syslog', 'stackdriver', 'stream'.
        **kwargs (optional): Keyword arguments to pass to handlers. See
            handler documentation for more information on available
            kwargs.
    """
    for h in handlers:
        if h == 'stream':
            handler = _setup_default_handler(progname, **kwargs)
        else:
            handler_module_path = 'ulogger.{}'.format(h)
            try:
                handler_module = import_module(
                    handler_module_path, package='ulogger')
            except ImportError:
                msg = 'Unsupported log handler: "{}".'.format(h)
                raise exceptions.ULoggerError(msg)

            try:
                get_handler = getattr(handler_module, 'get_handler')
            except AttributeError:
                msg = '"get_handler" function not implemented for "{}".'
                raise exceptions.ULoggerError(msg.format(h))

            handler = get_handler(progname, **kwargs)
        logging.getLogger('').addHandler(handler)

    level = logging.getLevelName(level)
    logging.getLogger('').setLevel(level)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import sys
from types import ModuleType

import pytest

from ulogger import exceptions
from ulogger import ulogger


@pytest.fixture
def logging_mock(mocker, monkeypatch):
    logging_module = mocker.MagicMock(logging, autospec=True)
    monkeypatch.setattr(ulogger, 'logging', logging_module)
    return logging_module


@pytest.fixture
def import_module_mock(mocker, monkeypatch):
    import_module = mocker.MagicMock('importlib.import_module', autospec=True)
    module_mock = mocker.MagicMock()
    handler = mocker.MagicMock()
    module_mock.get_handler = mocker.MagicMock()
    module_mock.get_handler.return_value = handler
    import_module_mock.return_value = module_mock
    monkeypatch.setattr(ulogger, 'import_module', import_module)
    return import_module


@pytest.fixture
def setup_default_handler_mock(mocker, monkeypatch):
    default = mocker.MagicMock(logging.StreamHandler, autospec=True)
    monkeypatch.setattr(ulogger, '_setup_default_handler', default)
    return default


args = 'handler,level'
params = [
    ('syslog', 'INFO'),
    ('stream', 'DEBUG')
]


@pytest.mark.parametrize(args, params)
def test_setup_logging(handler, level, logging_mock, import_module_mock,
                       setup_default_handler_mock):
    ulogger.setup_logging('foo', level, [handler])

    if handler == 'stream':
        handler_mock = setup_default_handler_mock.return_value
    else:
        module_path = 'ulogger.{}'.format(handler)
        import_module_mock.assert_called_once_with(
            module_path, package='ulogger')
        handler_mock = import_module_mock.return_value.get_handler.return_value

    logging_mock.getLevelName.assert_called_once_with(level)
    logging_mock.getLogger.assert_called_with('')
    assert logging_mock.getLogger.call_count == 2

    get_logger = logging_mock.getLogger.return_value
    get_logger.addHandler.assert_called_once_with(handler_mock)
    get_logger.setLevel.assert_called_once_with(
        logging_mock.getLevelName(level))


def test_setup_logging_multiple_handlers(import_module_mock,
                                         setup_default_handler_mock):
    handlers = ['stream', 'syslog']
    ulogger.setup_logging('tests', 'INFO', handlers)
    logger = logging.getLogger('')
    actual_handlers = logger.handlers
    assert len(actual_handlers) == len(handlers)


def test_setup_logging_raises():
    with pytest.raises(exceptions.ULoggerError) as e:
        ulogger.setup_logging('tests', 'INFO', ['notahandler'])

    assert e.match('Unsupported log handler: "notahandler".')


def test_setup_logging_raises_not_implemented():
    module_name = 'fake_handler'
    # Create a 'virtual' module that lacks the expected function
    fake_module = ModuleType(module_name)
    sys.modules['ulogger.{}'.format(module_name)] = fake_module

    with pytest.raises(exceptions.ULoggerError) as e:
        ulogger.setup_logging('tests', 'INFO', [module_name])

    assert e.match(
        '"get_handler" function not implemented for "{}"'.format(module_name))


fmts = [
    # explicit default
    '%(asctime)s.%(msecs)03dZ foo (%(process)d) %(levelname)s: %(message)s',
    # custom
    '%(created)f %(levelno)d %(message)s',
    # implicit default
    None
]
datefmts = [
    # explicit default
    '%Y-%m-%dT%H:%M:%S',
    # custom
    '%d/%m/%y %H:%M:%S.%f',
    # implicit default
    None
]
args = [(f, d) for f in fmts for d in datefmts]
params = 'fmt,datefmt'


@pytest.mark.parametrize(params, args)
def test_setup_default_handler(mocker, monkeypatch, fmt, datefmt):
    handler_mock = mocker.MagicMock(logging.StreamHandler, autospec=True)
    monkeypatch.setattr(ulogger.logging, 'StreamHandler', handler_mock)
    logging_mock = mocker.MagicMock(logging, autospec=True)
    monkeypatch.setattr(ulogger, 'logging', logging_mock)

    ret_handler = ulogger._setup_default_handler('foo', fmt, datefmt)

    if not fmt:
        fmt = fmts[0]
    if not datefmt:
        datefmt = datefmts[0]

    logging_mock.Formatter.assert_called_once_with(
        fmt=fmt, datefmt=datefmt)

    ret_handler.setFormatter.assert_called_once_with(
        logging_mock.Formatter.return_value)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

import logging

import requests
from google.cloud import logging as gcl_logging
from google.cloud.logging import handlers as gcl_handlers
from google.cloud.logging import resource as gcl_resource

from ulogger import exceptions


class CloudLoggingHandlerBuilder:
    """Creates instances of `google.cloud.logging.handlers.CloudLoggingHandler`

    Instances of the handler will be configured by retrieving the host's
    metadata. It can send logs to a different GCP project than the host
    is in by providing 'project_id'.

    Example usage:

        import logging
        from ulogging import stackdriver

        logger = logging.getLogger('')
        gcl_handler = stackdriver.CloudLoggingHandlerBuilder().get_handler()
        logger.setLevel(logging.INFO)
        log.info('Test message')

    Args:
        progname (str): Name of program.
        fmt (:obj:`str`, optional): Desired log format if different than
            the default; uses the same formatting string options
            supported in the stdlib's `logging` module.
        datefmt (:obj:`str`, optional): Desired date format if different
            than the default; uses the same formatting string options
            supported in the stdlib's `logging` module.
        project_id (:obj:`str`, optional): Project under which logs will
            be saved in GCL. If not provided, will get it from host's
            metadata.
        credentials (:obj:`obj`, optional): An instance of `google.auth.
            credentials.Credentials`. If not provided, will infer from
            environment.
        debug_thread_worker (:obj:`bool`, optional): Whether the
            background logging thread should emit DEBUG messages. If
            `False`, thread logger level is set to INFO.
    """
    METADATA_ENDPOINT = ('http://metadata.google.internal/computeMetadata/v1/'
                         '{data_type}/{key}')

    def __init__(self,
                 progname,
                 fmt=None,
                 datefmt=None,
                 project_id=None,
                 credentials=None,
                 debug_thread_worker=False):
        if project_id:
            self.project_id = project_id
        else:
            self.project_id = self._get_metadata(
                data_type='project', key='project-id')

        self.progname = progname
        self.fmt = fmt
        self.datefmt = datefmt
        self.credentials = credentials
        self.debug_thread_worker = debug_thread_worker
        self.hostname = self._get_metadata(data_type='instance', key='name')
        self.instance_id = self._get_metadata(data_type='instance', key='id')
        zone_str = self._get_metadata(data_type='instance', key='zone')
        self.zone = zone_str.split('/')[-1]
        self.resource = self._create_gcl_resource()

    def _get_metadata(self, data_type, key, timeout=5):
        """Get host instance metadata (only works on GCP hosts).

        More details about instance metadata:
        https://cloud.google.com/compute/docs/storing-retrieving-metadata

        Args:
            data_type (str): Type of metadata to fetch. Eg. project,
                instance
            key (str): Key of metadata to fetch
            timeout (int, optional): HTTP request timeout in seconds.
                Default is 5 seconds.
        Returns:
            (str): Plain text value of metadata entry
        Raises:
            GoogleCloudError: when request to metadata endpoint fails
        """
        endpoint_url = self.METADATA_ENDPOINT.format(
            data_type=data_type, key=key)
        try:
            rsp = requests.get(
                endpoint_url,
                headers={'Metadata-Flavor': 'Google'},
                timeout=timeout)
            rsp.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise exceptions.GoogleCloudError(
                'Could not fetch "{key}" from "{type}" metadata using "{url}".'
                'Error: {e}'.format(
                    key=key, type=data_type, url=endpoint_url, e=e))
        metadata_value = rsp.text
        if metadata_value.strip() == '':
            raise exceptions.GoogleCloudError(
                'Error when fetching metadata from "{url}": server returned '
                'an empty value.'.format(url=endpoint_url))
        return metadata_value

    def _create_gcl_resource(self):
        """Create a configured Resource object.

        The logging.resource.Resource object enables GCL to filter and
        bucket incoming logs according to which resource (host) they're
        coming from.

        Returns:
            (obj): Instance of `google.cloud.logging.resource.Resource`
        """

        return gcl_resource.Resource('gce_instance', {
            'project_id': self.project_id,
            'instance_id': self.instance_id,
            'zone': self.zone
        })

    def get_formatter(self):
        """Create a fully configured `logging.Formatter`

        Example of formatted log message:
        2017-08-27T20:19:24.424 cpm-example-gew1 progname (23123): hello

        Returns:
            (obj): Instance of `logging.Formatter`
        """
        if not self.fmt:
            self.fmt = ('%(asctime)s.%(msecs)03d {host} {progname} '
                        '(%(process)d): %(message)s').format(
                        host=self.hostname, progname=self.progname)
        if not self.datefmt:
            self.datefmt = '%Y-%m-%dT%H:%M:%S'
        return logging.Formatter(fmt=self.fmt, datefmt=self.datefmt)

    def _set_worker_thread_level(self):
        """Sets logging level of the background logging thread to DEBUG or INFO
        """
        bthread_logger = logging.getLogger(
            'google.cloud.logging.handlers.transports.background_thread')
        if self.debug_thread_worker:
            bthread_logger.setLevel(logging.DEBUG)
        else:
            bthread_logger.setLevel(logging.INFO)

    def get_handler(self):
        """Create a fully configured CloudLoggingHandler.

        Returns:
            (obj): Instance of `google.cloud.logging.handlers.
                                CloudLoggingHandler`
        """

        gcl_client = gcl_logging.Client(
            project=self.project_id, credentials=self.credentials)
        handler = gcl_handlers.CloudLoggingHandler(
            gcl_client,
            resource=self.resource,
            labels={
                'resource_id': self.instance_id,
                'resource_project': self.project_id,
                'resource_zone': self.zone,
                'resource_host': self.hostname
            })
        handler.setFormatter(self.get_formatter())
        self._set_worker_thread_level()
        return handler


def get_handler(progname, fmt=None, datefmt=None, project_id=None,
                credentials=None, debug_thread_worker=False, **_):
    """Helper function to create a Stackdriver handler.

    See `ulogger.stackdriver.CloudLoggingHandlerBuilder` for arguments
    and supported keyword arguments.

    Returns:
        (obj): Instance of `google.cloud.logging.handlers.
                            CloudLoggingHandler`
    """
    builder = CloudLoggingHandlerBuilder(
        progname, fmt=fmt, datefmt=datefmt, project_id=project_id,
        credentials=credentials, debug_thread_worker=debug_thread_worker)
    return builder.get_handler()



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pytest
import requests

from ulogger import exceptions
from ulogger import stackdriver


@pytest.fixture
def mock_requests_get(mocker):
    mock = mocker.Mock()
    project_id_rsp = mocker.Mock()
    project_id_rsp.text = 'test-project'
    host_rsp = mocker.Mock()
    host_rsp.text = 'cpm-guc99-hostname-1a'
    instance_id_rsp = mocker.Mock()
    instance_id_rsp.text = '123123'
    zone_rsp = mocker.Mock()
    zone_rsp.text = 'projects/192472736367/zones/us-central1-f'
    # Ordering matters and reflects calls in CloudLoggingHandlerBuilder.__init__
    mock.side_effect = [
        project_id_rsp, host_rsp, instance_id_rsp, zone_rsp,
    ]
    mocker.patch('ulogger.stackdriver.requests.get', mock)
    return mock


@pytest.fixture
def mock_cloud_handler(mocker):
    mock = mocker.Mock(
        stackdriver.CloudLoggingHandlerBuilder, autospec=True)
    mocker.patch('ulogger.stackdriver.CloudLoggingHandlerBuilder', mock)
    return mock


@pytest.fixture
def mock_logging_resource(mocker):
    mock_resource = mocker.Mock()
    mocker.patch('ulogger.stackdriver.gcl_resource.Resource', mock_resource)
    return mock_resource


@pytest.fixture
def mock_get_logger(mocker):
    mock_logger = mocker.Mock()
    mock_get_logger = mocker.Mock(return_value=mock_logger)
    mocker.patch('ulogger.stackdriver.logging.getLogger', mock_get_logger)
    return mock_get_logger, mock_logger


@pytest.fixture
def mock_gcl_client(mocker, mock_get_logger):
    _, mock_logger = mock_get_logger
    mock_client = mocker.MagicMock()
    mock_client_class = mocker.Mock(return_value=mock_client)
    mocker.patch(
        'ulogger.stackdriver.gcl_logging.Client',
        mock_client_class)
    return mock_client, mock_client_class


@pytest.fixture
def mock_gcl_handler(mocker):
    mock_handler = mocker.Mock()
    mock_cloud_handler = mocker.Mock(
        stackdriver.gcl_handlers.CloudLoggingHandler,
        autospec=True, return_value=mock_handler)
    mocker.patch('ulogger.stackdriver.gcl_handlers.CloudLoggingHandler',
                 mock_cloud_handler)
    return mock_handler, mock_cloud_handler


def test_get_handler(mocker):
    builder = mocker.Mock()
    mocker.patch('ulogger.stackdriver.CloudLoggingHandlerBuilder', builder)

    stackdriver.get_handler(
        'test-progname', project_id='example-project', credentials=None)

    builder.assert_called_once_with(
            'test-progname', project_id='example-project', credentials=None,
            fmt=None, datefmt=None, debug_thread_worker=False)


def test_builder_create_gcl_resource(mocker, mock_requests_get,
                                     mock_logging_resource):
    stackdriver.CloudLoggingHandlerBuilder('test-progname')

    expected_args = ['gce_instance', {
        'project_id': 'test-project',
        'instance_id': '123123',
        'zone': 'us-central1-f'}]

    mock_logging_resource.assert_called_once_with(*expected_args)


builder_kwargs = [
    {'project_id': None, 'credentials': None},
    {'project_id': 'example-project', 'credentials': 'example-credentials'}
]


@pytest.mark.parametrize('project_id,credentials', builder_kwargs)
def test_builder_get_handler(mocker, mock_requests_get, mock_logging_resource,
                             project_id, credentials, mock_get_logger,
                             mock_gcl_client, mock_gcl_handler):

    mock_client, mock_client_class = mock_gcl_client
    mock_handler, mock_cloud_handler = mock_gcl_handler
    mock_formatter = mocker.Mock()
    mocker.patch('ulogger.stackdriver.CloudLoggingHandlerBuilder.get_formatter',
                 mocker.Mock(return_value=mock_formatter))

    factory = stackdriver.CloudLoggingHandlerBuilder(
            'test-program', project_id=project_id, credentials=credentials)
    factory.get_handler()

    expected_labels = {
        'resource_id': factory.instance_id,
        'resource_project': factory.project_id,
        'resource_zone': factory.zone,
        'resource_host': factory.hostname}
    mock_cloud_handler.assert_called_once_with(
        mock_client,
        resource=factory.resource,
        labels=expected_labels)
    expected_project_id = project_id if project_id else factory.project_id
    mock_client_class.assert_called_once_with(
        project=expected_project_id, credentials=credentials)
    mock_handler.setFormatter.assert_called_once_with(mock_formatter)
    get_logger, logger = mock_get_logger
    get_logger.assert_called_once_with(
        'google.cloud.logging.handlers.transports.background_thread')
    logger.setLevel.assert_called_once_with(logging.INFO)


def test_builder_get_metadata(mock_requests_get):
    gapi_endpoint = ('http://metadata.google.internal/computeMetadata'
                     '/v1/{0}/{1}')
    stackdriver.CloudLoggingHandlerBuilder('test-progname')

    expected_gapi_headers = {'Metadata-Flavor': 'Google'}
    expected_url_params = [
        ('instance', 'id'),
        ('instance', 'zone'),
        ('project', 'project-id'),
    ]
    for data_type, key in expected_url_params:
        mock_requests_get.assert_any_call(
            gapi_endpoint.format(data_type, key),
            headers=expected_gapi_headers,
            timeout=5)


def test_builder_get_metadata_network_error(mock_requests_get):
    mock_requests_get.side_effect = requests.exceptions.RequestException(
        'Network error!')

    with pytest.raises(exceptions.GoogleCloudError):
        stackdriver.CloudLoggingHandlerBuilder('test-progname')


def test_builder_get_metadata_raises_on_empty_rsp(mocker, mock_requests_get):
    mock_rsp = mocker.Mock()
    mock_rsp.text = ''
    mock_requests_get.side_effect = [mock_rsp]

    with pytest.raises(exceptions.GoogleCloudError):
        stackdriver.CloudLoggingHandlerBuilder('test-progname')


fmts = [
    # explicit default
    ('%(asctime)s.%(msecs)03d cpm-guc99-hostname-1a test-progname '
     '(%(process)d): %(message)s'),
    # custom
    '%(created)f %(levelno)d %(message)s',
    # implicit default
    None
]
datefmts = [
    # explicit default
    '%Y-%m-%dT%H:%M:%S',
    # custom
    '%d/%m/%y %H:%M:%S.%f',
    # implicit default
    None
]
args = [(f, d) for f in fmts for d in datefmts]
params = 'fmt,datefmt'


@pytest.mark.parametrize(params, args)
def test_builder_get_formatter(mocker, mock_requests_get, fmt, datefmt):
    builder = stackdriver.CloudLoggingHandlerBuilder(
        'test-progname', fmt, datefmt)
    formatter = builder.get_formatter()

    if not fmt:
        fmt = fmts[0]
    if not datefmt:
        datefmt = datefmts[0]

    assert formatter._fmt == fmt
    assert formatter.datefmt == datefmt


def test_set_worker_thread_level_to_debug(mock_requests_get, mock_get_logger,
                                          mock_gcl_handler, mock_gcl_client):
    get_logger, logger = mock_get_logger
    stackdriver.CloudLoggingHandlerBuilder(
        'test-progname', debug_thread_worker=True).get_handler()

    get_logger.assert_called_once_with(
        'google.cloud.logging.handlers.transports.background_thread')
    logger.setLevel.assert_called_once_with(logging.DEBUG)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from ulogger.ulogger import setup_logging

__author__ = 'Lynn Root'
__version__ = '1.0.0'
__license__ = 'Apache 2.0'
__email__ = 'lynn@spotify.com'
__description__ = 'Micro logging library'
__uri__ = 'https://github.com/spotify/ulogger'

__all__ = ['setup_logging']



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import logging.handlers

import pytest

from ulogger import syslog


EXP_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'
SYSLOG_DEFAULT_FMT = '%(asctime)s.%(msecs)03dZ foo (%(process)d): %(message)s'
SYSLOG_OSX_FMT = 'foo (%(process)d): %(message)s'
REMOTE_ENV = {
    'SYSLOG_HOST': 'localhost',
    'SYSLOG_PORT': '12345'
}


def test_get_handler(mocker, monkeypatch):
    builder = mocker.MagicMock()
    monkeypatch.setattr('ulogger.syslog.SyslogHandlerBuilder', builder)

    syslog.get_handler('foo')

    builder.assert_called_once_with(
        'foo', address=None, proto=None, facility=None, fmt=None, datefmt=None)
    builder.return_value.get_handler.assert_called_once_with()


@pytest.fixture
def syslog_mock(mocker, monkeypatch):
    syslog_mock = mocker.MagicMock(
        logging.handlers.SysLogHandler, autospec=True)
    monkeypatch.setattr(syslog.logging.handlers, 'SysLogHandler', syslog_mock)


params = [
    # linux env
    ['default', ((syslog.sys, 'platform', 'linux2'),),
     '/dev/log', SYSLOG_DEFAULT_FMT, EXP_DATE_FORMAT],
    # remote/docker/helios env
    ['remote', ((syslog.os, 'environ', REMOTE_ENV),),
     ('localhost', 12345), SYSLOG_DEFAULT_FMT, EXP_DATE_FORMAT],
    # os x env
    ['darwin', ((syslog.sys, 'platform', 'darwin'),
                (syslog.os.path, 'exists', lambda x: True)),
     ('/var/run/syslog'), SYSLOG_OSX_FMT, None]
]
args = 'environ,patches,address,fmt,datefmt'


@pytest.mark.parametrize(args, params)
def test_syslog_handler_builder_env(environ, patches, address, fmt, datefmt,
                                    syslog_mock, mocker, monkeypatch):
    for patch in patches:
        monkeypatch.setattr(patch[0], patch[1], patch[2])

    builder = syslog.SyslogHandlerBuilder('foo')

    assert builder._environ == environ
    assert builder.address == address

    formatter = builder.get_formatter()
    assert formatter._fmt == fmt
    assert formatter.datefmt == datefmt

    formatter_mock = mocker.MagicMock(logging.Formatter, autospec=True)
    monkeypatch.setattr(syslog.logging, 'Formatter', formatter_mock)

    handler = builder.get_handler()
    handler.setFormatter.assert_called_once_with(formatter_mock.return_value)


fmts = [
    # explicit default
    '%(asctime)s.%(msecs)03dZ foo (%(process)d): %(message)s',
    # custom
    '%(created)f %(levelno)d %(message)s',
    # implicit default
    None
]
datefmts = [
    # explicit default
    '%Y-%m-%dT%H:%M:%S',
    # custom
    '%d/%m/%y %H:%M:%S.%f',
    # implicit default
    None
]
args = [(f, d) for f in fmts for d in datefmts]
params = 'fmt,datefmt'


@pytest.mark.parametrize(params, args)
def test_syslog_handler_builder_fmts(fmt, datefmt, monkeypatch):
    monkeypatch.setattr(syslog.sys, 'platform', 'linux2')
    builder = syslog.SyslogHandlerBuilder('foo', fmt=fmt, datefmt=datefmt)

    if not fmt:
        fmt = fmts[0]
    if not datefmt:
        datefmt = datefmts[0]

    formatter = builder.get_formatter()
    assert formatter._fmt == fmt
    assert formatter.datefmt == datefmt


args = [
    (None, 16),
    (None, logging.handlers.SysLogHandler.LOG_LOCAL0),
    (1, 1),
    (logging.handlers.SysLogHandler.LOG_USER, 1),
]

params = 'given,exp'


@pytest.mark.parametrize(params, args)
def test_syslog_handler_builder_facility(given, exp, mocker, monkeypatch):
    monkeypatch.setattr(syslog.sys, 'platform', 'linux2')

    with mocker.patch('socket.socket') as mock_socket:
        mock_socket.return_value.recv.return_value = True
        builder = syslog.SyslogHandlerBuilder('foo', facility=given)

    handler = builder.get_handler()

    assert builder.facility == exp
    assert handler.facility == exp


args = [
    # ((given address), given proto), (exp address), exp proto)
    ((('localhost', 514), 2), (('localhost', 514), 2)),
    ((('localhost', 514), None), (('localhost', 514), 2)),
    ((('10.99.0.1', None), 1), (('10.99.0.1', 514), 1)),
    (('/dev/log', None), ('/dev/log', 2)),
]

params = 'given,exp'


@pytest.mark.parametrize(params, args)
def test_syslog_handler_builder_address(given, exp, mocker, monkeypatch):
    monkeypatch.setattr(syslog.sys, 'platform', 'linux2')

    with mocker.patch('socket.socket') as mock_socket:
        mock_socket.return_value.recv.return_value = True
        builder = syslog.SyslogHandlerBuilder(
            'foo', address=given[0], proto=given[1])

    handler = builder.get_handler()

    assert builder.address == exp[0]
    assert builder.proto == exp[1]
    assert handler.address == exp[0]
    assert handler.socktype == exp[1]


args = [
    (
        (('localhost', 514), 2),  # explicit address/proto set
        (None, None),  # no envs set
        (('localhost', 514), 2)  # expected address/socktype
    ), (
        (None, None),  # no given address/proto
        ('localhost', None),  # host env set
        (('localhost', 514), 2)  # expected address/socktype
    ), (
        (None, None),  # no given address/proto
        ('localhost', 514),  # host and port envs set
        (('localhost', 514), 2)  # expected address/socktype
    ), (
        (None, 1),  # explicit proto set
        ('localhost', 514),  # host and port envs set
        (('localhost', 514), 1)  # expected address/socktype
    ), (
        (('localhost', 514), 2),  # explicit address/proto set
        ('127.0.0.1', 515),  # host and port envs set
        (('localhost', 514), 2)  # expected address/socktype
    )

]

params = 'given,envs,exp'


@pytest.mark.parametrize(params, args)
def test_syslog_handler_builder_address_envs(given, envs, exp,
                                             monkeypatch, mocker):
    monkeypatch.setattr(syslog.sys, 'platform', 'linux2')

    patch = {}
    if envs[0]:
        patch['SYSLOG_HOST'] = envs[0]
    if envs[1]:
        patch['SYSLOG_PORT'] = envs[1]

    if patch:
        monkeypatch.setattr(syslog.os, 'environ', patch)

    with mocker.patch('socket.socket') as mock_socket:
        mock_socket.return_value.recv.return_value = True
        builder = syslog.SyslogHandlerBuilder(
            'foo', address=given[0], proto=given[1])

    handler = builder.get_handler()

    assert builder.address == exp[0]
    assert builder.proto == exp[1]
    assert handler.address == exp[0]
    assert handler.socktype == exp[1]



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import logging.handlers
import os
import sys


class SyslogHandlerBuilder:
    """Creates a Syslog handler based on current environment.

    Example usage:

        import logging
        from ulogger import syslog

        logger = logging.getLogger('')
        syslog_handler = syslog.get_handler('foo_program')
        logger.addHandler(syslog_handler)

    Args:
        progname (str): Name of program.
        address (:obj:`tuple(str, int)` or :obj:`str`, optional):
            Address to send logs as either (host, port) or filepath.
            Defaults to `/dev/log` on Linux, and `/var/run/syslog` on
            OS X. Builder will also pick up environment variables
            `SYSLOG_HOST` and `SYSLOG_PORT`, but will prefer explicitly
            given addresses.
        proto (:obj:`int`, optional): Protocol for the address if
            `(host, port)` is given. Options: `1` for TCP
            (`socket.SOCK_STREAM`), or `2` for UDP (`socket.SOCK_DGRAM`).
            Defaults to `2`.
        facility (:obj:`int`, optional): desired facility with which to
            log. Can be either an `int` or a constant from
            `logging.handlers.SysLogHandler`. Default is `16` (local0).
        fmt (:obj:`str`, optional): Desired log format if different than
            the default; uses the same formatting string options
            supported in the stdlib's `logging` module.
        datefmt (:obj:`str`, optional): Desired date format if different
            than the default; uses the same formatting string options
            supported in the stdlib's `logging` module.
    """

    FACILITY = logging.handlers.SysLogHandler.LOG_LOCAL0

    def __init__(self, progname, address=None, proto=None, facility=None,
                 fmt=None, datefmt=None):
        self.progname = progname
        self._environ = self._get_environ()
        self.fmt = fmt
        self.datefmt = datefmt
        self.facility = facility or self.FACILITY
        self.address = self._get_address(address)
        self.proto = proto or 2  # 2 = socket.SOCK_DGRAM

    def _get_environ(self):
        syslog_host = os.environ.get('SYSLOG_HOST', None)
        if syslog_host:
            return 'remote'
        if (sys.platform.startswith('darwin') and
                os.path.exists('/var/run/syslog')):
            return 'darwin'
        return 'default'

    def _get_address(self, address):
        if address:
            if len(address) == 2:
                if address[1] is None:
                    return (address[0], logging.handlers.SYSLOG_UDP_PORT)
            return address
        address = {
            'default': '/dev/log',
            'darwin': '/var/run/syslog',
            'remote': (os.environ.get('SYSLOG_HOST'),
                       os.environ.get('SYSLOG_PORT',
                                      logging.handlers.SYSLOG_UDP_PORT))
        }[self._environ]
        if self._environ == 'remote':
            return (address[0], int(address[1]))
        return address

    def _get_osx_formatter(self):
        if not self.fmt:
            # MMM DD HH:MM:SS <host> <progname> (<PID>): <msg>
            # ex: Aug 25 13:00:51 foo.example.com bar (16911): hello
            self.fmt = self.progname + ' (%(process)d): %(message)s'
        return logging.Formatter(fmt=self.fmt)

    def _get_default_formatter(self):
        if not self.fmt:
            # e.g. 2017-08-25T14:47:44.968+00:00 <host> <progname>[<PID>]: <msg>
            prefix = '%(asctime)s.%(msecs)03dZ '
            suffix = ' (%(process)d): %(message)s'

            self.fmt = prefix + self.progname + suffix

        if not self.datefmt:
            self.datefmt = '%Y-%m-%dT%H:%M:%S'
        return logging.Formatter(fmt=self.fmt, datefmt=self.datefmt)

    def get_formatter(self):
        if self._environ == 'darwin':
            formatter = self._get_osx_formatter()
        else:
            formatter = self._get_default_formatter()
        return formatter

    def get_handler(self):
        handler = logging.handlers.SysLogHandler(
            address=self.address, facility=self.facility, socktype=self.proto)
        formatter = self.get_formatter()
        handler.setFormatter(formatter)
        return handler


def get_handler(progname, address=None, proto=None, facility=None,
                fmt=None, datefmt=None, **_):
    """Helper function to create a Syslog handler.

    See `ulogger.syslog.SyslogHandlerBuilder` for arguments and
    supported keyword arguments.

    Returns:
        (obj): Instance of `logging.SysLogHandler`
    """
    builder = SyslogHandlerBuilder(
        progname, address=address, proto=proto, facility=facility,
        fmt=fmt, datefmt=datefmt)
    return builder.get_handler()
